/***********************************************************************************
                         anselm.ru [2015-02-26] GNU GPL
***********************************************************************************/
#include "sax.h"
//#include "log.h"
#include <stdio.h>
#include <unistd.h>
#include <string.h> //strlen

#if defined(__amigaos__) && defined(__USE_INLINE__)
#include <proto/expat.h>
#endif

#ifdef XML_LARGE_SIZE
#if defined(XML_USE_MSC_EXTENSIONS) && _MSC_VER < 1400
#define XML_FMT_INT_MOD "I64"
#else
#define XML_FMT_INT_MOD "ll"
#endif
#else
#define XML_FMT_INT_MOD "l"
#endif

#include <iostream>
using namespace std;

void XMLCALL
startElement(void *data, const char *el, const char **attr) {
  SAX *sax = (SAX*) data;
  
  sax->d_stack.push(el);
  if(sax->d_depth==0) {
    sax->d_state = SAX::START_DOCUMENT;
    sax->doc_start();
    //if(!sax->d_header.empty()) sax->header(sax->d_header.c_str());
  }
  (sax->d_depth)++;
  
  sax->tag_start(el);
  //printf("<%s %d>\n", el, sax->d_depth);
  for (int i = 0; attr[i]; i += 2) {
    //printf(" %s='%s'", attr[i], attr[i + 1]);
    sax->attr(attr[i], attr[i + 1]);
    //(*xml)[ attr[i] ] = attr[i+1];
  }
  //printf("\n");
}

void XMLCALL
endElement(void *data, const char *el) {
  SAX *sax = (SAX*) data;
  sax->tag_end(el);
  
  (sax->d_depth)--;
  if(sax->d_depth==0) {
  	sax->d_state = SAX::END_DOCUMENT;
  	sax->doc_end();
  }
  //printf("</%s %d>\n", el, sax->d_depth);
}

void XMLCALL
characterData(void *data, const XML_Char *s, int len) {
  SAX *xml = (SAX*) data;
  
  SAX *sax = (SAX*) data;
  if(!sax->d_stack.empty()) {
    string value = s;
    value.resize(len);
    sax->trim(value);
    string name = sax->d_stack.top();
    sax->d_stack.pop();
    sax->tag(name.c_str(), value.c_str());
  }
}

SAX::SAX() {
  d_parser = XML_ParserCreate(NULL);
  if (! d_parser) {
    //warning("error: couldn't allocate memory for parser");
    exit(-1);
  }
  
  reset();
}

SAX::~SAX() {
  XML_ParserFree(d_parser);
}


SAX&
SAX::operator<< (const string& s) {
  parse(s.c_str());
  return *this;
}

bool
SAX::parse(const char* buf) { // прежде, чем вызвать expat, надо причесать xml, т.к expat капризен
  if(d_state==END_DOCUMENT) {
    reset();
  }
  
  if(d_state==NONE_DOCUMENT) {
    size_t len = strcspn(buf, "<");
    d_header.append(buf, len);
    buf+=len;
    
    size_t pos = strspn(buf, "<");
    if(pos==0) {
      return true; // нечего парсить
    }
    else { // есть, что парсить
      if(!d_header.empty()) {
        header(d_header.c_str()); // заголовок сформирован полностью
        d_header.clear();
      }
    }
  }
  
  string Buf = d_cache.append(buf, strlen(buf));
  d_cache.clear();
  
  //if(Buf.find("<")!=string::npos || Buf.find(">")!=string::npos) {
  if(Buf.find_first_of("<>")!=string::npos) {
    d_state=START_DOCUMENT;//обязательно надо вручную установить, что документ начался
  }
  
  while(!Buf.empty()) { // попасть в цикл можно только избавившись от заголовка

    size_t i = Buf.find(">");
    if(i==string::npos) { // завершения тэга ещё не пришло
      d_cache = Buf;
      return true;
    }
    
    string s = Buf.substr(0, i+1);
    
    if(!parse_expat(s.c_str(), s.size())) {
      if(d_state==END_DOCUMENT) { // такое считаем допустимым после окончания документа - т.к. можно ожидать следующий SOAP-заголовок
        reset();
        return parse(Buf.c_str());
      }
      //warning("error: mailformd XML");
      return false;
    }
    Buf.erase(0, i+1);
  }
  
  return true;
}

bool
SAX::parse_expat(const char* buf, size_t len) {
  return XML_Parse(d_parser, buf, len, 0); // 0 - означает, что XML не закончен и ожидается новый кусок
}

void
SAX::reset() {
  XML_ParserReset(d_parser, NULL);
  XML_SetUserData(d_parser, this);
  XML_SetElementHandler(d_parser, startElement, endElement);
  XML_SetCharacterDataHandler(d_parser, characterData);
  
  while(!d_stack.empty()) d_stack.pop();
  d_header.clear();
  d_depth=0;
  d_state = NONE_DOCUMENT;
  d_cache.clear();
}

int
SAX::depth() const {
  return d_depth;
}

void
SAX::trim(string& s) const {
  const char* t = (char*)" \t\n\r\f";
  s.erase(0, s.find_first_not_of(t));
  s.erase(s.find_last_not_of(t)+1);
}

XML_Index
SAX::ind() const {
  return XML_GetCurrentByteIndex(d_parser);
}

XML_Size
SAX::col() const {
  return  XML_GetCurrentColumnNumber(d_parser);
}

XML_Size
SAX::line() const {
  return XML_GetCurrentLineNumber(d_parser);
}

void
SAX::tag(const char* name, const char* value) {
  for(int i=0; i<d_depth; i++) printf(" ");
  printf("%s==%s\n", name, value);
}

void
SAX::attr(const char* name, const char* value) {
  tag(name, value);
}

void
SAX::header(const char* content) {
  printf("--------------HEADER--------------\n");
  printf("%s\n", content);
  printf("----------------------------------\n");
}
